import * as changeCase from 'change-case';
import dayjs from 'dayjs';
import fs from 'fs-extra';
import got from 'got';
import path from 'path';
import prettier from 'prettier';
import { castArray, cloneDeepFast, dedent, groupBy, isEmpty, isFunction, last, memoize, noop, omit, uniq, values } from 'vtils';
import { exec } from 'child_process';
import { getCachedPrettierOptions, getNormalizedRelativePath, getRequestDataJsonSchema, getResponseDataJsonSchema, jsonSchemaToType, sortByWeights, throwError } from './utils';
import { SwaggerToYApiServer } from './SwaggerToYApiServer';
export class Generator {
  /** 配置 */
  constructor(config, options = {
    cwd: process.cwd()
  }) {
    this.options = options;
    this.config = [];
    this.disposes = [];
    this.fetchProject = memoize(async ({
      serverUrl,
      token
    }) => {
      const projectInfo = await this.fetchApi(`${serverUrl}/api/project/get`, {
        token: token
      });
      const basePath = `/${projectInfo.basepath || '/'}`.replace(/\/+$/, '').replace(/^\/+/, '/');
      projectInfo.basepath = basePath; // 实现项目在 YApi 上的地址

      projectInfo._url = `${serverUrl}/project/${projectInfo._id}/interface/api`;
      return projectInfo;
    }, ({
      serverUrl,
      token
    }) => `${serverUrl}|${token}`);
    this.fetchExport = memoize(async ({
      serverUrl,
      token
    }) => {
      const projectInfo = await this.fetchProject({
        serverUrl,
        token
      });
      const categoryList = await this.fetchApi(`${serverUrl}/api/plugin/export`, {
        type: 'json',
        status: 'all',
        isWiki: 'false',
        token: token
      });
      return categoryList.map(cat => {
        var _cat$list, _cat$list$, _cat$list2, _cat$list2$;

        const projectId = ((_cat$list = cat.list) == null ? void 0 : (_cat$list$ = _cat$list[0]) == null ? void 0 : _cat$list$.project_id) || 0;
        const catId = ((_cat$list2 = cat.list) == null ? void 0 : (_cat$list2$ = _cat$list2[0]) == null ? void 0 : _cat$list2$.catid) || 0; // 实现分类在 YApi 上的地址

        cat._url = `${serverUrl}/project/${projectId}/interface/api/cat_${catId}`;
        cat.list = (cat.list || []).map(item => {
          const interfaceId = item._id; // 实现接口在 YApi 上的地址

          item._url = `${serverUrl}/project/${projectId}/interface/api/${interfaceId}`;
          item.path = `${projectInfo.basepath}${item.path}`;
          return item;
        });
        return cat;
      });
    }, ({
      serverUrl,
      token
    }) => `${serverUrl}|${token}`);
    // config 可能是对象或数组，统一为数组
    this.config = castArray(config);
  }

  async prepare() {
    this.config = await Promise.all( // config 可能是对象或数组，统一为数组
    this.config.map(async item => {
      if (item.serverType === 'swagger') {
        const swaggerToYApiServer = new SwaggerToYApiServer({
          swaggerJsonUrl: item.serverUrl
        });
        item.serverUrl = await swaggerToYApiServer.start();
        this.disposes.push(() => swaggerToYApiServer.stop());
      }

      if (item.serverUrl) {
        // 去除地址后面的 /
        // fix: https://github.com/fjc0k/yapi-to-typescript/issues/22
        item.serverUrl = item.serverUrl.replace(/\/+$/, '');
      }

      return item;
    }));
  }

  async generate() {
    const outputFileList = Object.create(null);
    await Promise.all(this.config.map(async (serverConfig, serverIndex) => {
      const projects = serverConfig.projects.reduce((projects, project) => {
        projects.push(...castArray(project.token).map(token => ({ ...project,
          token: token
        })));
        return projects;
      }, []);
      return Promise.all(projects.map(async (projectConfig, projectIndex) => {
        const projectInfo = await this.fetchProjectInfo({ ...serverConfig,
          ...projectConfig
        });
        await Promise.all(projectConfig.categories.map(async (categoryConfig, categoryIndex) => {
          // 分类处理
          // 数组化
          let categoryIds = castArray(categoryConfig.id); // 全部分类

          if (categoryIds.includes(0)) {
            categoryIds.push(...projectInfo.cats.map(cat => cat._id));
          } // 唯一化


          categoryIds = uniq(categoryIds); // 去掉被排除的分类

          const excludedCategoryIds = categoryIds.filter(id => id < 0).map(Math.abs);
          categoryIds = categoryIds.filter(id => !excludedCategoryIds.includes(Math.abs(id))); // 删除不存在的分类

          categoryIds = categoryIds.filter(id => !!projectInfo.cats.find(cat => cat._id === id)); // 顺序化

          categoryIds = categoryIds.sort();
          const codes = (await Promise.all(categoryIds.map(async (id, categoryIndex2) => {
            categoryConfig = { ...categoryConfig,
              id: id
            };
            const syntheticalConfig = { ...serverConfig,
              ...projectConfig,
              ...categoryConfig,
              mockUrl: projectInfo.getMockUrl()
            };
            syntheticalConfig.target = syntheticalConfig.target || 'typescript';
            syntheticalConfig.devUrl = projectInfo.getDevUrl(syntheticalConfig.devEnvName);
            syntheticalConfig.prodUrl = projectInfo.getProdUrl(syntheticalConfig.prodEnvName); // 接口列表

            let interfaceList = await this.fetchInterfaceList(syntheticalConfig);
            interfaceList = interfaceList.map(interfaceInfo => {
              // 实现 _project 字段
              interfaceInfo._project = omit(projectInfo, ['cats', 'getMockUrl', 'getDevUrl', 'getProdUrl']); // 预处理

              const _interfaceInfo = isFunction(syntheticalConfig.preproccessInterface) ? syntheticalConfig.preproccessInterface(cloneDeepFast(interfaceInfo), changeCase) : interfaceInfo;

              return _interfaceInfo;
            }).filter(Boolean);
            interfaceList.sort((a, b) => a._id - b._id);
            const interfaceCodes = await Promise.all(interfaceList.map(async interfaceInfo => {
              const outputFilePath = path.resolve(this.options.cwd, typeof syntheticalConfig.outputFilePath === 'function' ? syntheticalConfig.outputFilePath(interfaceInfo, changeCase) : syntheticalConfig.outputFilePath);
              const categoryUID = `_${serverIndex}_${projectIndex}_${categoryIndex}_${categoryIndex2}`;
              const code = await this.generateInterfaceCode(syntheticalConfig, interfaceInfo);
              const weights = [serverIndex, projectIndex, categoryIndex, categoryIndex2];
              return {
                categoryUID,
                outputFilePath,
                weights,
                code
              };
            }));
            const groupedInterfaceCodes = groupBy(interfaceCodes, item => item.outputFilePath);
            return Object.keys(groupedInterfaceCodes).map(outputFilePath => {
              const categoryCode = [...sortByWeights(groupedInterfaceCodes[outputFilePath]).map(item => item.code)].filter(Boolean).join('\n\n');

              if (!outputFileList[outputFilePath]) {
                outputFileList[outputFilePath] = {
                  syntheticalConfig,
                  content: [],
                  requestFunctionFilePath: syntheticalConfig.requestFunctionFilePath ? path.resolve(this.options.cwd, syntheticalConfig.requestFunctionFilePath) : path.join(path.dirname(outputFilePath), 'request.ts'),
                  requestHookMakerFilePath: syntheticalConfig.reactHooks && syntheticalConfig.reactHooks.enabled ? syntheticalConfig.reactHooks.requestHookMakerFilePath ? path.resolve(this.options.cwd, syntheticalConfig.reactHooks.requestHookMakerFilePath) : path.join(path.dirname(outputFilePath), 'makeRequestHook.ts') : ''
                };
              }

              return {
                outputFilePath: outputFilePath,
                code: categoryCode,
                weights: last(sortByWeights(groupedInterfaceCodes[outputFilePath])).weights
              };
            });
          }))).flat();

          for (const groupedCodes of values(groupBy(codes, item => item.outputFilePath))) {
            sortByWeights(groupedCodes);
            outputFileList[groupedCodes[0].outputFilePath].content.push(...groupedCodes.map(item => item.code));
          }
        }));
      }));
    }));
    return outputFileList;
  }

  async write(outputFileList) {
    return Promise.all(Object.keys(outputFileList).map(async outputFilePath => {
      let {
        // eslint-disable-next-line prefer-const
        content,
        requestFunctionFilePath,
        requestHookMakerFilePath,
        // eslint-disable-next-line prefer-const
        syntheticalConfig
      } = outputFileList[outputFilePath];
      const rawRequestFunctionFilePath = requestFunctionFilePath;
      const rawRequestHookMakerFilePath = requestHookMakerFilePath; // 支持 .jsx? 后缀

      outputFilePath = outputFilePath.replace(/\.js(x)?$/, '.ts$1');
      requestFunctionFilePath = requestFunctionFilePath.replace(/\.js(x)?$/, '.ts$1');
      requestHookMakerFilePath = requestHookMakerFilePath.replace(/\.js(x)?$/, '.ts$1'); //

      syntheticalConfig.typesOnly = true;

      if (!syntheticalConfig.typesOnly) {
        if (!(await fs.pathExists(rawRequestFunctionFilePath))) {
          await fs.outputFile(requestFunctionFilePath, dedent`
                import { RequestFunctionParams } from 'yapi-to-typescript'

                export interface RequestOptions {
                  /**
                   * 使用的服务器。
                   *
                   * - \`prod\`: 生产服务器
                   * - \`dev\`: 测试服务器
                   * - \`mock\`: 模拟服务器
                   *
                   * @default prod
                   */
                  server?: 'prod' | 'dev' | 'mock',
                }

                export default function request<TResponseData>(
                  payload: RequestFunctionParams,
                  options: RequestOptions = {
                    server: 'prod',
                  },
                ): Promise<TResponseData> {
                  return new Promise<TResponseData>((resolve, reject) => {
                    // 基本地址
                    const baseUrl = options.server === 'mock'
                      ? payload.mockUrl
                      : options.server === 'dev'
                        ? payload.devUrl
                        : payload.prodUrl

                    // 请求地址
                    const url = \`\${baseUrl}\${payload.path}\`

                    // 具体请求逻辑
                  })
                }
              `);
        }

        if (syntheticalConfig.reactHooks && syntheticalConfig.reactHooks.enabled && !(await fs.pathExists(rawRequestHookMakerFilePath))) {
          await fs.outputFile(requestHookMakerFilePath, dedent`
                import { useState, useEffect } from 'react'
                import { RequestConfig } from 'yapi-to-typescript'
                import { Request } from ${JSON.stringify(getNormalizedRelativePath(requestHookMakerFilePath, outputFilePath))}
                import baseRequest from ${JSON.stringify(getNormalizedRelativePath(requestHookMakerFilePath, requestFunctionFilePath))}

                export default function makeRequestHook<TRequestData, TRequestConfig extends RequestConfig, TRequestResult extends ReturnType<typeof baseRequest>>(request: Request<TRequestData, TRequestConfig, TRequestResult>) {
                  type Data = TRequestResult extends Promise<infer R> ? R : TRequestResult
                  return function useRequest(requestData: TRequestData) {
                    // 一个简单的 Hook 实现，实际项目可结合其他库使用，比如：
                    // @umijs/hooks 的 useRequest (https://github.com/umijs/hooks)
                    // swr (https://github.com/zeit/swr)

                    const [loading, setLoading] = useState(true)
                    const [data, setData] = useState<Data>()

                    useEffect(() => {
                      request(requestData).then(data => {
                        setLoading(false)
                        setData(data as any)
                      })
                    }, [JSON.stringify(requestData)])

                    return {
                      loading,
                      data,
                    }
                  }
                }
              `);
        }
      } // 始终写入主文件


      const rawOutputContent = dedent`
          /* tslint:disable */
          /* eslint-disable */

          /* 该文件由 yapi-to-typescript 自动生成，请勿直接修改！！！ */

          ${syntheticalConfig.typesOnly ? dedent`
                // @ts-ignore
                type FileData = File

                ${content.join('\n\n').trim()}
              ` : dedent`
                // @ts-ignore
                // prettier-ignore
                import { Method, RequestBodyType, ResponseBodyType, RequestConfig, RequestFunctionRestArgs, FileData, prepare } from 'yapi-to-typescript'
                // @ts-ignore
                import request from ${JSON.stringify(getNormalizedRelativePath(outputFilePath, requestFunctionFilePath))}
                ${content.join('\n\n').trim()}
              `}
        `; // ref: https://prettier.io/docs/en/options.html

      const prettyOutputContent = prettier.format(rawOutputContent, { ...(await getCachedPrettierOptions()),
        filepath: outputFilePath
      });
      const outputContent = `${dedent`
          /* prettier-ignore-start */
          ${prettyOutputContent}
          /* prettier-ignore-end */
        `}\n`;
      await fs.outputFile(outputFilePath, outputContent); // 如果要生成 JavaScript 代码，
      // 则先对主文件进行 tsc 编译，主文件引用到的其他文件也会被编译，
      // 然后，删除原始的 .tsx? 文件。

      if (syntheticalConfig.target === 'javascript') {
        await this.tsc(outputFilePath);
        await Promise.all([fs.remove(requestFunctionFilePath).catch(noop), fs.remove(requestHookMakerFilePath).catch(noop), fs.remove(outputFilePath).catch(noop)]);
      }
    }));
  }

  async tsc(file) {
    return new Promise(resolve => {
      // add this to fix bug that not-generator-file-on-window
      const command = `${require('os').platform() === 'win32' ? 'node ' : ''}${require.resolve(`typescript/bin/tsc`)}`;
      exec(`${command} --target ES2019 --module ESNext --jsx preserve --declaration --esModuleInterop ${file}`, {
        cwd: this.options.cwd,
        env: process.env
      }, () => resolve());
    });
  }

  async fetchApi(url, query) {
    const {
      body: res
    } = await got.get(url, {
      searchParams: query,
      responseType: 'json',
      https: {
        rejectUnauthorized: false
      }
    });
    /* istanbul ignore next */

    if (res && res.errcode) {
      throwError(res.errmsg);
    }

    return res.data || res;
  }

  /** 获取分类的接口列表 */
  async fetchInterfaceList({
    serverUrl,
    token,
    id
  }) {
    const category = ((await this.fetchExport({
      serverUrl,
      token
    })) || []).find(cat => !isEmpty(cat) && !isEmpty(cat.list) && cat.list[0].catid === id);

    if (category) {
      category.list.forEach(interfaceInfo => {
        // 实现 _category 字段
        interfaceInfo._category = omit(category, ['list']);
      });
    }

    return category ? category.list : [];
  }
  /** 获取项目信息 */


  async fetchProjectInfo(syntheticalConfig) {
    const projectInfo = await this.fetchProject(syntheticalConfig);
    const projectCats = await this.fetchApi(`${syntheticalConfig.serverUrl}/api/interface/getCatMenu`, {
      token: syntheticalConfig.token,
      project_id: projectInfo._id
    });
    return { ...projectInfo,
      cats: projectCats,
      getMockUrl: () => `${syntheticalConfig.serverUrl}/mock/${projectInfo._id}`,
      getDevUrl: devEnvName => {
        const env = projectInfo.env.find(e => e.name === devEnvName);
        return env && env.domain ||
        /* istanbul ignore next */
        '';
      },
      getProdUrl: prodEnvName => {
        const env = projectInfo.env.find(e => e.name === prodEnvName);
        return env && env.domain ||
        /* istanbul ignore next */
        '';
      }
    };
  }
  /** 生成接口代码 */


  async generateInterfaceCode(syntheticalConfig, interfaceInfo) {
    const extendedInterfaceInfo = { ...interfaceInfo,
      parsedPath: path.parse(interfaceInfo.path)
    };
    const requestFunctionName = isFunction(syntheticalConfig.getRequestFunctionName) ? await syntheticalConfig.getRequestFunctionName(extendedInterfaceInfo, changeCase) : changeCase.camelCase(extendedInterfaceInfo.parsedPath.name);
    const requestDataTypeName = isFunction(syntheticalConfig.getRequestDataTypeName) ? await syntheticalConfig.getRequestDataTypeName(extendedInterfaceInfo, changeCase) : changeCase.pascalCase(`${requestFunctionName}Request`);
    const responseDataTypeName = isFunction(syntheticalConfig.getResponseDataTypeName) ? await syntheticalConfig.getResponseDataTypeName(extendedInterfaceInfo, changeCase) : changeCase.pascalCase(`${requestFunctionName}Response`);
    const requestDataJsonSchema = getRequestDataJsonSchema(extendedInterfaceInfo, syntheticalConfig.customTypeMapping || {});
    const requestDataType = await jsonSchemaToType(requestDataJsonSchema, requestDataTypeName);
    const responseDataJsonSchema = getResponseDataJsonSchema(extendedInterfaceInfo, syntheticalConfig.customTypeMapping || {}, syntheticalConfig.dataKey);
    const responseDataType = await jsonSchemaToType(responseDataJsonSchema, responseDataTypeName);
    const isRequestDataOptional = /(\{\}|any)$/s.test(requestDataType); // 接口注释

    const genComment = (genTitle, isCommentEnabled) => {
      const {
        enabled: isEnabled = isCommentEnabled,
        title: hasTitle = true,
        category: hasCategory = false,
        tag: hasTag = false,
        requestHeader: hasRequestHeader = false,
        updateTime: hasUpdateTime = true,
        link: hasLink = true,
        extraTags
      } = { ...syntheticalConfig.comment,
        // Swagger 时总是禁用标签、更新时间、链接
        ...(syntheticalConfig.serverType === 'swagger' ? {
          tag: false,
          updateTime: false,
          link: false
        } : {})
      };

      if (!isEnabled) {
        return '';
      } // 转义标题中的 /


      const escapedTitle = String(extendedInterfaceInfo.title).replace(/\//g, '\\/');
      const description = hasLink ? `[${escapedTitle}↗](${extendedInterfaceInfo._url})` : escapedTitle;
      const summary = [hasCategory && {
        label: '分类',
        value: hasLink ? `[${extendedInterfaceInfo._category.name}↗](${extendedInterfaceInfo._category._url})` : extendedInterfaceInfo._category.name
      }, hasTag && {
        label: '标签',
        value: extendedInterfaceInfo.tag.map(tag => `\`${tag}\``)
      }, hasRequestHeader && {
        label: '请求头',
        value: `\`${extendedInterfaceInfo.method.toUpperCase()} ${extendedInterfaceInfo.path}\``
      }, hasUpdateTime && {
        label: '更新时间',
        value: process.env.JEST_WORKER_ID // 测试时使用 unix 时间戳
        ? String(extendedInterfaceInfo.up_time) :
        /* istanbul ignore next */
        `\`${dayjs(extendedInterfaceInfo.up_time * 1000).format('YYYY-MM-DD HH:mm:ss')}\``
      }];

      if (typeof extraTags === 'function') {
        const tags = extraTags(extendedInterfaceInfo);

        for (const tag of tags) {
          ;
          (tag.position === 'start' ? summary.unshift : summary.push).call(summary, {
            label: tag.name,
            value: tag.value
          });
        }
      }

      const titleComment = hasTitle ? dedent`
            * ${genTitle(description)}
          ` : '';
      const extraComment = summary.filter(item => typeof item !== 'boolean' && !isEmpty(item.value)).map(item => {
        const _item = item;
        return `* @${_item.label} ${castArray(_item.value).join(', ')}`;
      }).join('\n');
      return dedent`
        /**
         ${[titleComment, extraComment].filter(Boolean).join('\n')}
         */
      `;
    };

    const escapedTitle = String(extendedInterfaceInfo.title).replace(/\//g, '\\/');
    return dedent`
    /**
     * [${escapedTitle}] **请求类型**
     */
      ${requestDataType.trim()}

      /**
       * [${escapedTitle}] **返回类型**
       */
      ${responseDataType.trim()}

      ${syntheticalConfig.typesOnly ? '' : dedent`
            ${genComment(title => `接口 ${title}`, true)}
            export const ${requestFunctionName} = (requestData${isRequestDataOptional ? '?' : ''}: ${requestDataTypeName}):Promise<${responseDataTypeName}> => {
              return axios.${extendedInterfaceInfo.method.toLocaleLowerCase()}('${extendedInterfaceInfo.path}', { params: requestData })
            }
          `}
    `;
  }

  async destroy() {
    return Promise.all(this.disposes.map(async dispose => dispose()));
  }

}